import logging

from zygoat.components import SettingsComponent

log = logging.getLogger()


class InstalledApps(SettingsComponent):
    def create(self):
        exported_values = ["ALLOWED_HOSTS", "DATABASES"]

        red = self.parse()
        apps_list = red.find("name", value="INSTALLED_APPS").parent.value

        log.info("Adding DRF to installed apps")
        apps_list.append("'rest_framework'")

        log.info("Adding backend app to installed apps")
        apps_list.append("'zygoat_django'")

        log.info("Removing default components exported by zygoat_django")

        for value in exported_values:
            index = red.find("name", value=value).parent.index_on_parent

            # Delete the blank line following it as well
            del red[index]
            del red[index]

        log.info("Dumping installed apps node")
        self.dump(red)

    @property
    def installed(self):
        red = self.parse()
        return (
            "zygoat_django"
            in red.find("name", value="INSTALLED_APPS").parent.value.to_python()
        )


installed_apps = InstalledApps()



import logging
import os
import shutil

from zygoat.components import Component
from zygoat.constants import Projects, Phases
from zygoat.utils.shell import run
from zygoat.utils.files import use_dir

from .dockerfile import dockerfile
from .gitignore import gitignore
from .prettier import prettier
from .eslint import eslint
from .cypress import cypress
from .next_config import next_config
from .dependencies import dependencies
from .jsconfig import js_config

log = logging.getLogger()


class Frontend(Component):
    def create(self):
        log.info("Installing/upgrading yarn through npm")
        run(["npm", "install", "-g", "--upgrade", "yarn"])

        log.info("Running create-next-app")
        run(["yarn", "create", "next-app", Projects.FRONTEND])

        log.info("Emptying a poorly formatted index.js file")
        open(os.path.join(Projects.FRONTEND, "pages", "index.js"), "w").close()

        log.info("Deleting default _app.js file (added in create-next-app v9.5.5)")
        os.remove(os.path.join(Projects.FRONTEND, "pages", "_app.js"))

        log.info("Deleting default next.config.js (added in create-next-app v11)")
        os.remove(os.path.join(Projects.FRONTEND, "next.config.js"))

        log.info("Deleting the default api directory")
        shutil.rmtree(os.path.join(Projects.FRONTEND, "pages", "api"))

    def delete(self):
        log.warning(f"Deleting the {Projects.FRONTEND} project")
        shutil.rmtree(Projects.FRONTEND)

    @property
    def installed(self):
        return os.path.exists(Projects.FRONTEND)


class Reformat(Component):
    def create(self):
        log.info("Formatting package.json with prettier")
        with use_dir(Projects.FRONTEND):
            run(["yarn", "prettier", "-w", "package.json"])

    def update(self):
        self.call_phase(Phases.CREATE, force_create=True)

    @property
    def installed(self):
        # json.dump does not do a good job of preserving formatting
        with open(os.path.join(Projects.FRONTEND, "package.json")) as f:
            return "\n" in f.read()


frontend = Frontend(
    sub_components=[
        dockerfile,
        gitignore,
        prettier,
        dependencies,
        eslint,
        cypress,
        next_config,
        js_config,
        Reformat(),
    ]
)



import logging

from zygoat.components import SettingsComponent
from redbaron import RedBaron

from .installed_apps import installed_apps

log = logging.getLogger()


class Settings(SettingsComponent):
    def create(self):
        # RedBaron has a *really* obtuse bug where it will not insert end of line comments
        # but it will preserve them in an output dump - so we just insert the raw string here
        lines = self.parse().dumps().split("\n")

        for idx, line in enumerate(lines):
            if "import" in line:
                first_import_index = idx
                break

        log.info("Creating import for zygoat-django settings")
        lines.insert(first_import_index, "from zygoat_django.settings import *  # noqa")

        # We run it through RedBaron again to make sure we've generated valid source code
        red = RedBaron("\n".join(lines))
        self.dump(red)

    @property
    def installed(self):
        red = self.parse()

        return red.find("name", value="zygoat_django") is not None


settings_sub_components = [
    installed_apps,
]

settings = Settings(sub_components=settings_sub_components)



import os
from shutil import which

from .shell import run
from .files import use_dir, repository_root
from zygoat.constants import Projects

pip = which("pip")
dev_file_name = "requirements.dev.txt"
prod_file_name = "requirements.txt"


def freeze():
    return run([pip, "freeze"], capture_output=True).stdout.decode().split("\n")


def packages_to_map(arr):
    result = {}

    for package_line in arr:
        # Ignore `-r requirements.txt` and other non-package related lines
        if "=" not in package_line:
            if "git://" in package_line:
                result[package_line] = package_line
            continue
        if "[" in package_line:
            package = package_line.split("[")[0]
        else:
            package = package_line.split("=")[0]
        result[package] = package_line

    return result


def dump_dependencies(package_map, dev=False):
    file_name = dev_file_name if dev else prod_file_name

    with repository_root():
        with use_dir(Projects.BACKEND):
            with open(file_name, "w") as f:
                if dev:
                    f.write(f"-r {prod_file_name}\n")

                for name, version in package_map.items():
                    # Arbitrary vertical whitespace comes out at as emptystring, so ignore it
                    if name == "":
                        continue

                    f.write(f"{version}\n")


def install_dependencies(*args, dev=False, extras={}):
    """
    Installs/upgrades Python dependencies for the backend, and places them in
    the appropriate production or dev requirements files.

    :param args: The packages to install
    :type args: str
    :param dev: Specifies if this is a development or production dependency
    :type dev: bool, optional
    """
    initialize_files()
    file_name = dev_file_name if dev else prod_file_name
    with repository_root():
        with use_dir(Projects.BACKEND):
            pip_args = [
                "{}{}".format(
                    arg,
                    "[{}]".format(",".join(extras.get(arg))) if extras.get(arg) else "",
                )
                for arg in args
            ]

            run([pip, "install", "--upgrade", *pip_args])
            freeze_map = packages_to_map(freeze())

            with open(file_name) as f:
                file_map = packages_to_map(f.read().split("\n"))

            for name in args:
                file_map[name] = freeze_map[name]

                if extras.get(name):
                    ver = file_map[name].replace(
                        name, "{}[{}]".format(name, ",".join(extras.get(name)))
                    )
                    file_map[name] = ver

            dump_dependencies(file_map, dev=dev)


def initialize_files():
    with repository_root():
        with use_dir(Projects.BACKEND):
            if not os.path.exists(prod_file_name):
                open(prod_file_name, "w").close()

            if not os.path.exists(dev_file_name):
                open(dev_file_name, "w").close()



import click
from importlib import import_module
import logging
import semver

from .components import components
from .config import Config
from .constants import Phases, config_file_name, __version__

log = logging.getLogger()


class ZygoatOutOfDateException(Exception):
    """
    Exception raised when trying to run zygoat on a project initialized with
    a newer version of zygoat.
    """

    def __init__(self, current_version, original_version):
        super().__init__(
            f"Tried to update a zygoat project using a older "
            f"version than the project was created with."
            f"Run with --force if this is intended.\n"
            f"Original version: {original_version}\n"
            f"Current version: {current_version}"
        )


def _is_older_version(version1, version2):
    return semver.compare(version1, version2) < 0


def _call_phase(phase, reverse=False, force=False):
    component_list = [*components]
    config = Config()

    version = config.get("version")
    if _is_older_version(__version__, version):
        log.warn(
            f"Using older version of zygoat than when this package was made. "
            f"Current: {__version__}, Original: {version}."
        )
        if not force:
            raise ZygoatOutOfDateException(__version__, version)

    for extra in config.get("extras", []):
        module, attr = extra.split(":")

        log.info(f"Adding {extra} to the component tree")
        component_list.append(getattr(import_module(module), attr))

    if reverse:
        component_list.reverse()

    for component in component_list:
        component.call_phase(phase)


def _name_project(project_name):
    config = Config()

    if project_name == "" and getattr(config, "name", None) is not None:
        return

    config.name = project_name
    Config.dump(config)


@click.group()
@click.option("--verbose", "-v", is_flag=True)
def cli(verbose):
    if verbose:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)


@cli.command(help="Creates a new zygoat settings file and exits")
@click.argument("project_name")
def init(project_name):
    _name_project(project_name)

    log.info(f"Initialized {config_file_name}")


@cli.command(help="Creates a new settings file and all components")
@click.option("--force", "-f", is_flag=True)
@click.argument("project_name", default="")
def new(project_name, force):
    log.debug(f"Attempting creation of {click.style(project_name, bold=True)}")

    _name_project(project_name)

    _call_phase(Phases.CREATE, force=force)

    log.info("Done!")
    log.info(
        "Run {} to start your new {} project!".format(
            click.style("docker-compose up --build", fg="cyan", bold=True),
            click.style(project_name, fg="cyan", bold=True),
        )
    )


@cli.command(help="Lists all of the running phase names")
@click.option("--force", "-f", is_flag=True)
def list(force):
    _call_phase(Phases.LIST, force=force)


@cli.command(help="Calls the delete phase on all included build components")
@click.option("--force", "-f", is_flag=True)
def delete(force):
    _call_phase(Phases.DELETE, reverse=True, force=force)

    # remove zygoat settings file
    Config.delete()


@cli.command(help="Calls the update phase on all included build components")
@click.option("--force", "-f", is_flag=True)
def update(force):
    _call_phase(Phases.UPDATE, force=force)


@cli.command(help="Prints the version and exits")
def version():
    log.info(f"Running zygoat v{__version__}")



from zygoat.components import FileComponent, Component
from zygoat.constants import Projects

from . import resources


class NextConfigFile(FileComponent):
    resource_pkg = resources
    base_path = Projects.FRONTEND
    filename = "next.config.js"
    overwrite = False


class ZygoatNextConfigFile(FileComponent):
    resource_pkg = resources
    base_path = Projects.FRONTEND
    filename = "zygoat.next.config.js"
    overwrite = True


class BabelConfigFile(FileComponent):
    resource_pkg = resources
    base_path = Projects.FRONTEND
    filename = ".babelrc"
    overwrite = False


class NextConfig(Component):
    pass


next_config = NextConfig(
    sub_components=[NextConfigFile(), ZygoatNextConfigFile(), BabelConfigFile()]
)
